module Stlc2


exception ListError
exception TypeError
exception RunTimeError
exception Unimplemented (* shows you what to change *)

type var = string

type typ =
    Int
  | Arrow  of typ * typ (* Arrow(t1,t2) means function from t1 to t2 *)

type context = (var * typ) list

type exp =
  | Const of int
  | Plus of exp * exp
  | Var of var
  | Lam of var * typ * exp
  | Apply of exp ref * exp
  | Closure of exp * env (* not in source programs *)
and env = (var * exp) list

let empty = []

let extend lst x e = (x,e)::lst

let rec lookup lst y =
  match lst with
    [] -> raise ListError
  | (x,v)::tl -> if x=y then v else lookup tl y

let rec interpret_env env e =
  match e with
    Const i -> e
  | Plus(e1,e2)  ->
      let v1 = interpret_env env e1 in
      let v2 = interpret_env env e2 in
      (match (v1,v2) with
        Const i1, Const i2 -> Const(i1 + i2)
      |        _ -> raise RunTimeError)
  | Var(x) -> lookup env x
  | Lam _ -> Closure(e,env)
  | Apply(e1,e2) ->
      let v1 = interpret_env env !e1 in
      let v2 = interpret_env env e2 in
      (match v1 with
        Closure(Lam(x,_,e3),envc) -> interpret_env (extend envc x v2) e3
      |        _ -> raise RunTimeError)
  | Closure(_,_) -> e

let rec typecheck_ctxt ctxt e =
  match e with
    Const i -> Int
  | Plus(e1,e2) ->
      let t1 = typecheck_ctxt ctxt e1 in
      let t2 = typecheck_ctxt ctxt e2 in
      if (t1=Int) && (t2=Int)
      then Int
      else raise TypeError
  | Var(x) -> (try lookup ctxt x with ListError -> raise TypeError)
  | Lam(x,t,e) ->
      let t1 = typecheck_ctxt (extend ctxt x t) e in
      Arrow(t,t1)
  | Apply(e1,e2) ->
      let t1 = typecheck_ctxt ctxt !e1 in
      let t2 = typecheck_ctxt ctxt e2 in
      (match t1 with
        Arrow(t3,t4) -> if t3=t2 then t4 else raise TypeError
      |        _ -> raise TypeError)
  | Closure(_,_) -> raise TypeError

let interpret e = interpret_env empty e

let typecheck e = ignore(typecheck_ctxt empty e)

let interpret1 (e: exp) : exp = raise Unimplemented

let typecheck2 (e: exp) : unit = raise Unimplemented

let interpret2 (e : exp) : exp = raise Unimplemented

let typecheck3 (e : exp) : (unit -> exp) = raise Unimplemented

type checkedexp = Checked of exp

let typecheck4 (e : exp) : checkedexp = raise Unimplemented

let interpret4 (e: checkedexp) : exp = raise Unimplemented
